"""
author: arun.rs
21st September 2018
"""
import uuid

class TraceRequest(object):
    """
    Maintains a UID for every application request
    """
    __trace_id = ""
    __parent_id = ""
    __child_id = ""

    def __init__(self):
        pass

    def get_trace_id(self, renew=False):
        """
        :Brief: This method is used in every place to get the already generated request ID or
        generate new request ID and sent off
        """
        if not TraceRequest.__trace_id or renew:
            self.set_trace_id(str(uuid.uuid1()))
        return TraceRequest.__trace_id + "|" + TraceRequest.__parent_id + "|" + TraceRequest.__child_id

    def set_trace_id(self, trace_id, parent_id="", child_id=""):
        """
        :Brief: This method is used by entry points where the incoming HTTP request has the
        request ID in its headers
        """
        TraceRequest.__trace_id = trace_id
        if parent_id:
            TraceRequest.__parent_id = parent_id
        else:
            TraceRequest.__parent_id = trace_id
        TraceRequest.__child_id = child_id

    def get_root_id(self):
        """
        :summary:
        :return:
        """
        return TraceRequest.__trace_id

    def get_parent_id(self):
        """

        :return:
        """
        return TraceRequest.__parent_id

    def get_child_id(self):
        """
        :summary:
        :return:
        """
        return TraceRequest.__child_id

    def get_all_ids(self):
        """

        :return:
        """
        if not TraceRequest.__trace_id:
            self.set_trace_id(str(uuid.uuid1()))
        return TraceRequest.__trace_id, TraceRequest.__parent_id, TraceRequest.__child_id



from api import service
from dependency_injection import DependencyInjection
from logger import CustomLogger
from mysqldb_wrapper import MySQLdbWrapper
from trace import trace, trace_resource
from trace_request import TraceRequest
from utilities import sanitize_urlparams


"""
arun.rs
6th December 2018
"""
import re

def sanitize_urlparams(url):
    """
    :summary: Replace url params with <params> string
    :param url: URL
    :return: Sanitized URL
    """
    return re.sub("\=(.*?)(\&|$)", "=<params>&", url)



"""
author: arun.rs
24th September 2018
"""
import MySQLdb
from trace import trace_resource
from datetime import datetime

class MySQLdbWrapper(object):
    """
    MySQLdb Library wrapper
    """
    def __init__(self, db_config, cursor_type=None):
        """
        :summary: Reads the credentials from
        :param db_config: DB Config credentials
        :param cursor_type: Cursor Type
        """
        self.cursor_type = None
        self.db_config = db_config
        self.connection_obj = None
        self.cursor_obj = None
        self.set_cursor(cursor_type)

    def connect(self):
        """
        :summary: Created a MySQL Connection
        :return: None
        """
        try:
            status = True
            start_time = datetime.now()
            self.connection_obj = MySQLdb.connect(
                                                   host=self.db_config['host'],
                                                   user=self.db_config['user'],
                                                   passwd=self.db_config['passwd'],
                                                   db=self.db_config['db'],
                                                   port=self.db_config['port'],
                                                   connect_timeout=self.db_config['connect_timeout'],
                                                   charset=self.db_config['charset']
                                                   )
        except Exception as e:
            status = False
            raise e
        finally:
            end_time = datetime.now()
            trace_resource("mysql", "Connecting", "", "", start_time, end_time, status)

    def set_cursor(self, cursor_type):
        """
        :summary: Set cursor with the requested cursor_type
        :param cursor_type: Cursor Type
        :return: None
        """
        if (not self.connection_obj):
            self.connect()
        if cursor_type == "tuple":
            self.cursor_obj = self.connection_obj.cursor()
        else:
            self.cursor_obj = self.connection_obj.cursor(MySQLdb.cursors.DictCursor)

    def processquery(self, query, count=0, args=None, fetch=True, returnprikey=0):
        """
        :summary: Executes the Mysql query and returns the results
        :param query: SQL raw query
        :param count: Number of records to be returned
        :param args: Arguments to be passed to the query
        :param fetch: Fetch the query result or not (True / False)
        :param returnprikey: Return primary key if 1
        :return: Query result
        """
        status = True
        start_time = datetime.now()
        try:
            self.cursor_obj.execute(query, args)
            if fetch:
                if count == 1:
                    return self.cursor_obj.fetchone()
                elif count > 1:
                    return self.cursor_obj.fetchmany(count)
                else:
                    return self.cursor_obj.fetchall()
            else:
                if returnprikey:
                    return self.cursor_obj.lastrowid
                else:
                    return self.cursor_obj.rowcount
        except Exception as e:
            status = False
            raise e
        finally:
            end_time = datetime.now()
            trace_resource("mysql", "", "", query, start_time, end_time, status)

    def save(self):
        """
        :summary: Commits the transactions
        :return: None
        """
        self.connection_obj.commit()

    def close(self):
        """
        :summary: Closes the MySQL connection
        :return:None

        """
        self.connection_obj.close()

    def revert(self):
        """
        :summary: Roll back the transactions
        :return:None
        """
        self.connection_obj.rollback()



"""
author: arun.rs
created: 26th October 2018
"""

from datetime import datetime
from functools import wraps
from logger import CustomLogger

TRACER = CustomLogger().get_logger('trace')

def message(operation, type, resource, raw_resource, execution_time, status):
    """
    :summary: Concats the supplied parameters and returns them in trace format
    :param operation: Operation (MySQL/ Mongo/ ES/ API/ etc)
    :param type: Type of the Operation (SELECT/ GET/ POST/ etc)
    :param resource: URL / Query / Function name
    :param raw_resource: URL / Query / Function name
    :param execution_time: Time taken to perform that operation
    :param status: Success or Failure
    :return: Concatinated string
    """
    return "%s|%s|%s|%s|%s|%s" % (operation, type, resource, raw_resource, execution_time, status)

def execution_time(start_time, end_time):
    """
    :summary: Difference of supplied time in seconds
    :param start_time: Start time
    :param end_time: End time
    :return: Difference of supplied time in seconds
    """
    return str(((end_time-start_time).total_seconds()) * 1000)

def trace_resource(operation, type, resource, raw_resource, start_time, end_time, status):
    """
    :summary: Logs the information required to trace the application
    :param operation: Operation (Mysql/ Mongo/ ES/ API/ etc)
    :param type: Type of the Operation (GET/ POST/ SELECT/ INSERT/ etc)
    :param resource: URL / Query / Function name
    :param raw_resource: URL / Query / Function name
    :param start_time: Start time of the operation
    :param end_time: End time of the operation
    :param status: Success / Failure
    :return: None
    """
    TRACER.info(message(operation, type, resource, raw_resource, execution_time(start_time, end_time), status))

def trace(operation, type):
    """
    :summary: Wrapper function to trace a method
    :param operation: Operation (Mysql/ Mongo/ ES/ API/ etc)
    :param type: Type of the Operation (GET/ POST/ SELECT/ INSERT/ etc)
    :return: Traced functions response
    """
    def trace(func):
        @wraps(func)
        def trace(*args, **kwargs):
            start_time = datetime.now()
            status = True
            try:
                return func(*args, **kwargs)
            except Exception as e:
                status = False
            finally:
                end_time = datetime.now()
                trace_resource(operation, type, func.__name__, func.__name__, start_time, end_time, status)
        return trace
    return trace





"""
author: arun.rs
21st September 2018
"""
import os

from logging.config import dictConfig
#Logs absoulte location
LOG_PATH = '/logs/'
#Logs Generic settings
LOG_SETTINGS = {
    'version': 1,
    'disable_existing_loggers': False,

    'formatters': {
                    'standard': {'format': "%(asctime)s|%(levelname)s|%(module)s|%(lineno)s|%(funcName)s|%(message)s"},
                    'trace': {'format': "%(asctime)s|%(message)s"}
                  },

    'handlers': {
        'default': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
        }
    },

    'loggers': {
                }
}
#Module wise logs settings
LOG_SETTINGS['handlers']['application'] = {
    'class': 'cloghandler.ConcurrentRotatingFileHandler',
    'level': 'DEBUG',
    'formatter': 'standard',
    'filename': os.path.join(LOG_PATH, 'application.log'),
    'mode': 'a',
    'maxBytes': 104857600,
    'backupCount': 5,
}

LOG_SETTINGS['loggers']['application'] = {
    'handlers': [ 'application'],
    'formatter': 'standard',
    'level': 'DEBUG',
    'propagate': False
}

LOG_SETTINGS['handlers']['trace'] = {
    'class': 'cloghandler.ConcurrentRotatingFileHandler',
    'level': 'DEBUG',
    'formatter': 'trace',
    'filename': os.path.join(LOG_PATH, 'trace.log'),
    'mode': 'a',
    'maxBytes': 104857600,
    'backupCount': 5,
}

LOG_SETTINGS['loggers']['trace'] = {
    'handlers': ['trace'],
    'formatter': 'trace',
    'level': 'DEBUG',
    'propagate': False
}

dictConfig(LOG_SETTINGS)




"""
author: arun.rs
13th November 2018
"""

class DependencyInjection(object):
    """
    This class is used for sharing objects across the application.
    objects which are registered in this object will be shared across
    all the instances of this class
    """

    # All the services are stored in this class variable and available to all
    # the instances of di object
    services = {}

    def set(self, service_name, service_object):
        """
        :summary: This method will register the services in class variable
        @param service_name: Name for the service to register in DI. Supposed to be
                            unique for each service registered.
        :param service_object: The service object which needs to be registered.
        :return: Returns 1 if there is no service is already registered in the same given name,
        Else store the given service_object for the service_name and return 0.
        :rtype: int
        """
        # If a service is already registered in the same name, store in a variable or store in None.
        existing_value_in_service = DependencyInjection.services.get(service_name, None)
        DependencyInjection.services[service_name] = service_object
        # If already a service was registered in the same name return 0 else return 1
        if (existing_value_in_service):
            return 0
        else:
            return 1

    def get(self, service_name):
        """
        :summary: Retrieves the requested service and returns it.
        :param service_name: service name for which the objects need
        to be returned
        :raise ServiceNotFoundException: raised when the given service_name
        is not found in services
        """
        if DependencyInjection.services.get(service_name):
            return DependencyInjection.services.get(service_name)
        else:
            raise Exception("ServiceNotFound : " + service_name)


"""
author: arun.rs
26th September 2018
"""
from datetime import datetime
from requests import request
from trace import trace_resource
from trace_request import TraceRequest
from utilities import sanitize_urlparams

def service(type, url, **kwargs):
    """
    :summary: Wrapper method that traces all the HTTP requests made
    :param type: Service type
    :param url: URL to connect to
    :param kwargs: Key word arguments
    :return: Response
    """
    if not 'headers' in kwargs:
        headers = {}
    headers['trace_id'], headers['parent_id'], headers['child_id'] = TraceRequest().get_all_ids()
    status = True
    start_time = datetime.now()
    try:
        return request(type, url, **kwargs)
    except Exception as e:
        status = False
        raise
    finally:
        end_time = datetime.now()
        trace_resource("api", type, sanitize_urlparams(url), url, start_time, end_time, status)



from distutils.core import setup
setup(
  name = 'open_tracer',
  packages = ['open_tracer'],
  version = '2.0',
  license='MIT',
  description = 'Open-tracer : To trace every applications request',
  author = 'arun.rs',
  author_email = 'arunrs@uninstall.io',
  url = 'https://github.com/rsarun-uninstallio/open_tracer',
  download_url = 'https://github.com/rsarun-uninstallio/open_tracer/archive/v2.0.tar.gz',
  keywords = ['tracing', 'open_tracing', 'open_tracer'],
  install_requires=['uuid', 'requests', 'datetime', 'logging', 'mysqlclient', 'concurrentloghandler'],
  classifiers=[  # Optional
    # How mature is this project? Common values are
    #   3 - Alpha
    #   4 - Beta
    #   5 - Production/Stable
    'Development Status :: 3 - Alpha',

    # Indicate who your project is intended for
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',

    # Pick your license as you wish
    'License :: OSI Approved :: MIT License',

    # Specify the Python versions you support here. In particular, ensure
    # that you indicate whether you support Python 2, Python 3 or both.
    'Programming Language :: Python :: 2.7',
  ],
)


"""
author: arun.rs
created: 10th October 2018
"""
import logging.config
from trace_request import TraceRequest
from logger_settings import LOG_SETTINGS

class Logger(logging.LoggerAdapter):
    """
    :summary:
    """
    def __init__(self, logger):
        """
        :summary: Initiallizes the logger with the supplied log name
        :param logger: Logger name
        """
        self.logger = logger

    def process(self, msg, kwargs):
        """
        :summary: Overrides the default process method to append UUID to the logs
        :param msg: Message to be logger
        :param kwargs: Keyword arguments
        :return: Message with the request id and msg data
        """
        return '%s|%s' % (TraceRequest().get_trace_id(), msg), kwargs

class CustomLogger(object):
    """
    :summary: Custom logger with the supplied log settings
    """
    def __init__(self, log_setting=None):
        """
        :summary: Initiallizes logging module with the value passed, else sets default log settings
        :param log_setting: Logging settings
        :return: None
        """
        if log_setting:
            logging.config.dictConfig(log_setting)
        # else:
            # logging.basicConfig(filename='/logs/default.log', level=logging.DEBUG)
            # logging.config.dictConfig(LOG_SETTINGS)

    def get_logger(self, logger_val):
        """
        :summary: Validates the logger_instance and returns the same
        :param logger_val: Logger Value
        :return: Logger Instance
        """
        if isinstance(logger_val, str):
            logger_instance = logging.getLogger(logger_val)
        elif isinstance(logger_val, logging.LoggerAdapter):
            logger_instance = logger_val.logger
        elif isinstance(logger_val, logging.Logger):
            logger_instance = logger_val
        else:
            raise ValueError("Invalid Input")
        return Logger(logger_instance)

    def renew_request_id(self, value=None):
        """
        :summary: Renews the request id
        :param value: Sets the value as request id
        :return: None
        """
        if value:
            TraceRequest().set_trace_id(value)
        else:
            TraceRequest().get_trace_id(renew=True)
